from functools import lru_cache
from dataclasses import dataclass

from .model import Model


@dataclass
class PowerlawParams():
    modulus_0: float
    modulus_inf: float
    tau: float
    e: float

    def __post_init__(self):
        if self.modulus_0 <= 0:
            raise ValueError("modulus_0 must be greater than zero.")
        if self.modulus_inf <= 0:
            raise ValueError("modulus_inf must be greater than zero.")
        if self.tau <= 0:
            raise ValueError("tau must be greater than zero.")
        if not (0 <= self.e <= 1):
            raise ValueError("e must be between 0 and 1.")


class Powerlaw(Model):
    _params: PowerlawParams
    diagram = ""

    def __init__(self, params: PowerlawParams) -> None:
        super().__init__()
        self.params = params

    @property
    def params(self) -> PowerlawParams:
        return self._params

    @params.setter
    def params(self, params: PowerlawParams) -> None:
        if not isinstance(params, PowerlawParams):
            raise ValueError("Invalid parameters: Expected PowerlawParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.params.modulus_inf + (self.params.modulus_0 - self.params.modulus_inf) * (1 + t / self.params.tau) ** (-self.params.e)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        return 1 / self.params.modulus_inf + (1 / self.params.modulus_0) * (t / self.params.tau) ** self.params.e



from pyvisq import Test, TestMethod
from pyvisq.models import zener

# Define the test method and parameters
method = TestMethod.CREEP
test_params = {
    "I": 1.0,
    "D1": 0.01,
    "L1": 2,
    "D2": 0.01,
    "L2": 2
}
test = Test(method=method, **test_params)

# Define the Zener model parameters
dashpot_a = zener.DashpotParams(c=1)
spring_b = zener.SpringParams(k=1)
spring_c = zener.SpringParams(k=1)
sls_params = zener.SLSParams(
    dashpot_a=dashpot_a,
    spring_b=spring_b,
    spring_c=spring_c
)
sls = zener.SLS(params=sls_params)

# Print the SLS model diagram and parameters
print(sls)

# Set up and run the test
sls.set_test(test)
sls.set_time()
sls.set_input()  # Optional: set the input profile for visualization
sls.run()


import os
import ast
import pyvisq


def get_classes_from_file(file_path):
    with open(file_path, "r", encoding="utf-8") as f:
        node = ast.parse(f.read(), filename=file_path)
    classes = []
    for n in node.body:
        if isinstance(n, ast.ClassDef) and 'Params' not in n.name:
            classes.append(n.name)
    return classes


def generate_hierarchy_tree(root_dir):
    hierarchy = {}
    for dirpath, _, filenames in os.walk(root_dir):
        relative_path = os.path.relpath(dirpath, root_dir)
        if relative_path == ".":
            relative_path = ""

        for filename in filenames:
            if filename.endswith(".py") and filename != "model.py":
                file_path = os.path.join(dirpath, filename)
                class_list = get_classes_from_file(file_path)
                if class_list:
                    hierarchy[os.path.join(
                        relative_path, filename)] = class_list
    return hierarchy


def print_hierarchy_tree(hierarchy):
    for file, classes in sorted(hierarchy.items()):
        print(file[:-3])
        for cls in classes:
            print(f"  ├── {cls}")


def show():
    root_directory = os.path.dirname(pyvisq.__file__)
    hierarchy_tree = generate_hierarchy_tree(root_directory)
    print_hierarchy_tree(hierarchy_tree)



import mpmath

import numpy as np

from functools import lru_cache
from dataclasses import dataclass

from .model import Model
from .elements import Spring, Dashpot, Springpot, SpringParams, DashpotParams, SpringpotParams
from .maxwell import Maxwell, MaxwellParams, FracDashpotMaxwell, FracDashpotMaxwellParams, FracSpringMaxwell, FracSpringMaxwellParams, FracMaxwell, FracMaxwellParams


@dataclass
class SLSParams():
    dashpot_a: DashpotParams
    spring_b: SpringParams
    spring_c: SpringParams


class SLS(Model):
    _params: SLSParams
    dashpot_a: Dashpot
    spring_b: Spring
    maxwell_branch: Maxwell
    spring_c: Spring
    diagram = """
                    ___                 
                 ____| |______╱╲  ╱╲  ╱╲  _____
                |   _|_|  ca    ╲╱  ╲╱  ╲╱  kb |
            ____|                              |____
                |                              |
                |__________╱╲  ╱╲  ╱╲  ________|
                             ╲╱  ╲╱  ╲╱  kc
            """

    def __init__(self, params: SLSParams) -> None:
        super().__init__()
        self.params = params
        self.maxwell_branch = Maxwell(
            MaxwellParams(params.dashpot_a, params.spring_b)
        )
        self.dashpot_a = self.maxwell_branch.dashpot
        self.spring_b = self.maxwell_branch.spring
        self.spring_c = Spring(params.spring_c)

    @property
    def params(self) -> SLSParams:
        return self._params

    @params.setter
    def params(self, params: SLSParams) -> None:
        if not isinstance(params, SLSParams):
            raise ValueError("Invalid parameters: Expected SLSParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.spring_c.G(t) + self.maxwell_branch.G(t)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        ca = self.params.dashpot_a.c
        kb = self.params.spring_b.k
        kc = self.params.spring_c.k
        c0 = 1 / kc
        c1 = kb / (kc * (kc + kb))
        tau = ca * (kc + kb) / (kc * kb)
        return c0 - c1 * np.exp(-t / tau)


@dataclass
class JeffreysZenerParams():
    dashpot_a: DashpotParams
    spring_b: SpringParams
    dashpot_c: DashpotParams


class JeffreysZener(Model):
    _params: JeffreysZenerParams
    dashpot_a: Dashpot
    spring_b: Spring
    maxwell_branch: Maxwell
    dashpot_c: Dashpot
    diagram = """
                    ___                 
                 ____| |______╱╲  ╱╲  ╱╲  _____
                |   _|_|  ca    ╲╱  ╲╱  ╲╱  kb |
            ____|                              |____
                |              ___             |
                |_______________| |____________|
                               _|_|  cc
            """

    def __init__(self, params: JeffreysZenerParams) -> None:
        super().__init__()
        self.params = params
        self.maxwell_branch = Maxwell(
            MaxwellParams(params.dashpot_a, params.spring_b)
        )
        self.dashpot_a = self.maxwell_branch.dashpot
        self.spring_b = self.maxwell_branch.spring
        self.dashpot_c = Dashpot(params.dashpot_c)

    @property
    def params(self) -> JeffreysZenerParams:
        return self._params

    @params.setter
    def params(self, params: JeffreysZenerParams) -> None:
        if not isinstance(params, JeffreysZenerParams):
            raise ValueError(
                "Invalid parameters: Expected JeffreysZenerParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.dashpot_c.G(t) + self.maxwell_branch.G(t)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        if t == 0:
            return 0

        ca = self.params.dashpot_a.c
        kb = self.params.spring_b.k
        cc = self.params.dashpot_c.c

        def J_hat(s):
            numerator = ca * s + kb
            denominator = ca * kb * s + cc * s * (ca * s + kb)
            return (1 / s) * numerator / denominator

        t_mp = mpmath.mpf(t)
        J_t = mpmath.invertlaplace(J_hat, t_mp, method='talbot')
        return float(J_t)


@dataclass
class FracJeffreysZenerParams():
    dashpot_a: DashpotParams
    springpot_b: SpringpotParams
    dashpot_c: DashpotParams


class FracJeffreysZener(Model):
    _params: FracJeffreysZenerParams
    dashpot_a: Dashpot
    springpot_b: Springpot
    maxwell_branch: FracDashpotMaxwell
    dashpot_c: Dashpot
    diagram = """
                    ___                 
                 ____| |___________╱╲__________
                |   _|_|  ca       ╲╱  cb, b   |
            ____|                              |____
                |              ___             |
                |_______________| |____________|
                               _|_|  cc
            """

    def __init__(self, params: FracJeffreysZenerParams) -> None:
        super().__init__()
        self.params = params
        self.maxwell_branch = FracDashpotMaxwell(
            FracDashpotMaxwellParams(params.dashpot_a, params.springpot_b)
        )
        self.dashpot_a = self.maxwell_branch.dashpot
        self.springpot_b = self.maxwell_branch.springpot
        self.dashpot_c = Dashpot(params.dashpot_c)

    @property
    def params(self) -> FracJeffreysZenerParams:
        return self._params

    @params.setter
    def params(self, params: FracJeffreysZenerParams) -> None:
        if not isinstance(params, FracJeffreysZenerParams):
            raise ValueError(
                "Invalid parameters: Expected FracJeffreysZenerParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.dashpot_c.G(t) + self.maxwell_branch.G(t)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        if t == 0:
            return 0

        ca = self.params.dashpot_a.c
        b = self.params.springpot_b.e
        cb = self.params.springpot_b.ce
        cc = self.params.dashpot_c.c

        def J_hat(s):
            numerator = ca * s + cb * s ** b
            denominator = (ca * s) * (cb * s ** b) + \
                cc * s * (ca * s + cb * s ** b)
            return (1 / s) * numerator / denominator

        t_mp = mpmath.mpf(t)
        J_t = mpmath.invertlaplace(J_hat, t_mp, method='talbot')
        return float(J_t)


@dataclass
class FracSolidZenerParams():
    dashpot_a: DashpotParams
    springpot_b: SpringpotParams
    spring_c: SpringParams


class FracSolidZener(Model):
    _params: FracSolidZenerParams
    dashpot_a: Dashpot
    springpot_b: Springpot
    maxwell_branch: FracDashpotMaxwell
    spring_c: Spring
    diagram = """
                      ___                 
                 ______| |__________╱╲_________
                |     _|_|  ca      ╲╱  cb, b  |
            ____|                              |____
                |                              |
                |__________╱╲  ╱╲  ╱╲  ________|
                             ╲╱  ╲╱  ╲╱  kc
            """

    def __init__(self, params: FracSolidZenerParams) -> None:
        super().__init__()
        self.params = params
        self.maxwell_branch = FracDashpotMaxwell(
            FracDashpotMaxwellParams(params.dashpot_a, params.springpot_b)
        )
        self.dashpot_a = self.maxwell_branch.dashpot
        self.springpot_b = self.maxwell_branch.springpot
        self.spring_c = Spring(params.spring_c)

    @property
    def params(self) -> FracSolidZenerParams:
        return self._params

    @params.setter
    def params(self, params: FracSolidZenerParams) -> None:
        if not isinstance(params, FracSolidZenerParams):
            raise ValueError(
                "Invalid parameters: Expected FracSolidZenerParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.spring_c.G(t) + self.maxwell_branch.G(t)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        if t == 0:
            return 0

        ca = self.params.dashpot_a.c
        b = self.params.springpot_b.e
        cb = self.params.springpot_b.ce
        kc = self.params.spring_c.k

        def J_hat(s):
            A = ca / cb
            B = kc * A
            numerator = 1 + A * s ** (1 - b)
            denominator = ca + kc / s + B * s ** (-b)
            return (1 / s ** 2) * numerator / denominator

        t_mp = mpmath.mpf(t)
        J_t = mpmath.invertlaplace(J_hat, t_mp, method='talbot')
        return float(J_t)


@dataclass
class FracSLSZenerParams():
    springpot_a: SpringpotParams
    spring_b: SpringParams
    spring_c: SpringParams


class FracSLSZener(Model):
    _params: FracSLSZenerParams
    springpot_a: Springpot
    spring_b: Spring
    maxwell_branch: FracSpringMaxwell
    spring_c: Spring
    diagram = """                
                 ___╱╲________╱╲  ╱╲  ╱╲  _____
                |   ╲╱  ca, a   ╲╱  ╲╱  ╲╱  kb |
            ____|                              |____
                |                              |
                |__________╱╲  ╱╲  ╱╲  ________|
                             ╲╱  ╲╱  ╲╱  kc
            """

    def __init__(self, params: FracSLSZenerParams) -> None:
        super().__init__()
        self.params = params
        self.maxwell_branch = FracSpringMaxwell(
            FracSpringMaxwellParams(params.springpot_a, params.spring_b)
        )
        self.springpot_a = self.maxwell_branch.springpot
        self.spring_b = self.maxwell_branch.spring
        self.spring_c = Spring(params.spring_c)

    @property
    def params(self) -> FracSLSZenerParams:
        return self._params

    @params.setter
    def params(self, params: FracSLSZenerParams) -> None:
        if not isinstance(params, FracSLSZenerParams):
            raise ValueError(
                "Invalid parameters: Expected FracSLSZenerParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.spring_c.G(t) + self.maxwell_branch.G(t)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        if t == 0:
            return 0

        a = self.params.springpot_a.e
        ca = self.params.springpot_a.ce
        kb = self.params.spring_b.k
        kc = self.params.spring_c.k

        def J_hat(s):
            numerator = ca * s ** a + kb
            denominator = ca * kb * s ** a + kc * (ca * s ** a + kb)
            return (1 / s) * numerator / denominator

        t_mp = mpmath.mpf(t)
        J_t = mpmath.invertlaplace(J_hat, t_mp, method='talbot')
        return float(J_t)


@dataclass
class FracZenerParams():
    springpot_a: SpringpotParams
    springpot_b: SpringpotParams
    springpot_c: SpringpotParams


class FracZener(Model):
    _params: FracZenerParams
    springpot_a: Springpot
    springpot_b: Springpot
    maxwell_branch: FracMaxwell
    springpot_c: Springpot
    diagram = """               
                 ______╱╲___________╱╲_________
                |      ╲╱  ca, a    ╲╱  cb, b  |
            ____|                              |____
                |                              |
                |______________╱╲______________|
                               ╲╱  cc, c
            """

    def __init__(self, params: FracZenerParams) -> None:
        super().__init__()
        self.params = params
        self.maxwell_branch = FracMaxwell(
            FracMaxwellParams(params.springpot_a, params.springpot_b)
        )
        self.springpot_a = self.maxwell_branch.springpot_a
        self.springpot_b = self.maxwell_branch.springpot_b
        self.springpot_c = Springpot(params.springpot_c)

    @property
    def params(self) -> FracZenerParams:
        return self._params

    @params.setter
    def params(self, params: FracZenerParams) -> None:
        if not isinstance(params, FracZenerParams):
            raise ValueError("Invalid parameters: Expected FracZenerParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.springpot_c.G(t) + self.maxwell_branch.G(t)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        if t == 0:
            return 0

        a = self.params.springpot_a.e
        ca = self.params.springpot_a.ce
        b = self.params.springpot_a.e
        cb = self.params.springpot_b.ce
        c = self.params.springpot_c.e
        cc = self.params.springpot_c.ce

        def J_hat(s):
            numerator = ca * s ** a + cb * s ** b
            denominator = (ca * s ** a) * (cb * s ** b) + \
                (cc * s ** c) * (ca * s ** a + cb * s ** b)
            return (1 / s) * numerator / denominator

        t_mp = mpmath.mpf(t)
        J_t = mpmath.invertlaplace(J_hat, t_mp, method='talbot')
        return float(J_t)



import mpmath

from functools import lru_cache
from dataclasses import dataclass

from .model import Model
from .elements import Spring, SpringParams
from .maxwell import Maxwell, MaxwellParams


@dataclass
class GeneralizedParams():
    spring1: SpringParams
    branches: list[MaxwellParams]


class Generalized(Model):
    _params: GeneralizedParams
    spring1: Spring
    branches: list[Maxwell]
    diagram = """                
                 __________╱╲  ╱╲  ╱╲  ________
                |            ╲╱  ╲╱  ╲╱  k1    |
                |                              |
                |                     ___      |
            ____|__╱╲  ╱╲  ╱╲  ________| |_____|____
                |    ╲╱  ╲╱  ╲╱  k2   _|_|  c2 |
                :                :          :  :
                :                :    ___  :   : 
                |__╱╲  ╱╲  ╱╲  ________| |_____|
                     ╲╱  ╲╱  ╲╱  kn   _|_|  cn  
            """

    def __init__(self, params: GeneralizedParams) -> None:
        super().__init__()
        self.params = params
        self.spring1 = Spring(params.spring1)
        self.branches = [Maxwell(branch) for branch in params.branches]

    @property
    def params(self) -> GeneralizedParams:
        return self._params

    @params.setter
    def params(self, params: GeneralizedParams) -> None:
        if not isinstance(params, GeneralizedParams):
            raise ValueError("Invalid parameters: Expected GeneralizedParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.spring1.G(t) + sum([branch.G(t) for branch in self.branches])

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        if t == 0:
            return 0

        if len(self.branches) == 0:
            return self.spring1.J(t)

        def J_hat(s):
            all_branches = self.spring1.params.k / s
            for branch in self.branches:
                branch_params = branch.params
                k = branch_params.spring.k
                c = branch_params.dashpot.c
                tau = c / k
                all_branches += k / (s + 1 / tau)
            return (1 / s ** 2) / all_branches

        t_mp = mpmath.mpf(t)
        J_t = mpmath.invertlaplace(J_hat, t_mp, method='talbot')
        return float(J_t)



import numpy as np

from functools import lru_cache
from dataclasses import dataclass

from .model import Model
from .elements import Spring, Dashpot, Springpot, SpringParams, DashpotParams, SpringpotParams
from ..utils import MittagLeffler


@dataclass
class MaxwellParams():
    dashpot: DashpotParams
    spring: SpringParams


class Maxwell(Model):
    _params: MaxwellParams
    spring: Spring
    dashpot: Dashpot
    diagram = """
                ___
            _____| |_______╱╲  ╱╲  ╱╲  ___
                _|_|  c      ╲╱  ╲╱  ╲╱  k
            """

    def __init__(self, params: MaxwellParams) -> None:
        super().__init__()
        self.params = params
        self.dashpot = Dashpot(params.dashpot)
        self.spring = Spring(params.spring)

    @property
    def params(self) -> MaxwellParams:
        return self._params

    @params.setter
    def params(self, params: MaxwellParams) -> None:
        if not isinstance(params, MaxwellParams):
            raise ValueError("Invalid parameters: Expected MaxwellParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        c = self.params.dashpot.c
        k = self.params.spring.k
        tau = c / k
        return k * np.exp(-t / tau)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        return self.dashpot.J(t) + self.spring.J(t)


@dataclass
class FracDashpotMaxwellParams():
    dashpot: DashpotParams
    springpot: SpringpotParams


class FracDashpotMaxwell(Model):
    _params: FracDashpotMaxwellParams
    dashpot: Dashpot
    springpot: Springpot
    ml: MittagLeffler
    diagram = """
                ___
            _____| |_______╱╲____
                _|_|  c    ╲╱  cb, b
            """

    def __init__(self, params: FracDashpotMaxwellParams) -> None:
        super().__init__()
        self.params = params
        self.dashpot = Dashpot(params.dashpot)
        self.springpot = Springpot(params.springpot)
        self.ml = MittagLeffler()

    @property
    def params(self) -> FracDashpotMaxwellParams:
        return self._params

    @params.setter
    def params(self, params: FracDashpotMaxwellParams) -> None:
        if not isinstance(params, FracDashpotMaxwellParams):
            raise ValueError(
                "Invalid parameters: Expected FracDashpotMaxwellParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        t = max(t, 1e-10)
        c = self.params.dashpot.c
        b = self.params.springpot.e
        cb = self.params.springpot.ce
        return cb * t ** (-b) * self.ml(1 - b, 1 - b, -cb * t ** (1 - b) / c)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        return self.dashpot.J(t) + self.springpot.J(t)


@dataclass
class FracSpringMaxwellParams():
    springpot: SpringpotParams
    spring: SpringParams


class FracSpringMaxwell(Model):
    _params: FracSpringMaxwellParams
    springpot: Springpot
    spring: Spring
    ml: MittagLeffler
    diagram = """
            ____╱╲_________╱╲  ╱╲  ╱╲  ______
                ╲╱  ca, a    ╲╱  ╲╱  ╲╱  k
            """

    def __init__(self, params: FracSpringMaxwellParams) -> None:
        super().__init__()
        self.params = params
        self.spring = Spring(params.spring)
        self.springpot = Springpot(params.springpot)
        self.ml = MittagLeffler()

    @property
    def params(self) -> FracSpringMaxwellParams:
        return self._params

    @params.setter
    def params(self, params: FracSpringMaxwellParams) -> None:
        if not isinstance(params, FracSpringMaxwellParams):
            raise ValueError(
                "Invalid parameters: Expected FracSpringMaxwellParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        k = self.params.spring.k
        a = self.params.springpot.e
        ca = self.params.springpot.ce
        return k * self.ml(a, a, -k * t ** (a) / ca)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        return self.spring.J(t) + self.springpot.J(t)


@dataclass
class FracMaxwellParams():
    springpot_a: SpringpotParams
    springpot_b: SpringpotParams


class FracMaxwell(Model):
    _params: FracMaxwellParams
    springpot_a: Springpot
    springpot_b: Springpot
    ml: MittagLeffler
    diagram = """
            ____╱╲___________╱╲______
                ╲╱  ca, a    ╲╱  cb, b
            """

    def __init__(self, params: FracMaxwellParams) -> None:
        super().__init__()
        self.params = params
        self.springpot_a = Springpot(params.springpot_a)
        self.springpot_b = Springpot(params.springpot_b)
        self.ml = MittagLeffler()

    @property
    def params(self) -> FracMaxwellParams:
        return self._params

    @params.setter
    def params(self, params: FracMaxwellParams) -> None:
        if not isinstance(params, FracMaxwellParams):
            raise ValueError("Invalid parameters: Expected FracMaxwellParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        t = max(t, 1e-10)
        a = self.params.springpot_a.e
        ca = self.params.springpot_a.ce
        b = self.params.springpot_b.e
        cb = self.params.springpot_b.ce
        return cb * t ** (-b) * self.ml(a - b, 1 - b, -cb * t ** (a - b) / ca)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        return self.springpot_a.J(t) + self.springpot_b.J(t)



import numpy as np
import numpy.typing as npt

from typing import Callable
from functools import lru_cache
from scipy.integrate import quad
from enum import Enum
from dataclasses import dataclass, field


class TestMethod(Enum):
    CREEP = 'creep'
    RELAXATION = 'relaxation'


@dataclass
class Test:
    """
    ```
    input profile:
        I   ==========
          //:        :\\
         // :        : \\
        //  :D1      :L1\\D2==========L2
    ```
    """
    method: TestMethod
    I: float
    D1: float
    L1: float
    D2: float
    L2: float

    def __post_init__(self):
        if self.D1 <= 0:
            raise ValueError("D1 must be greater than zero.")
        if self.L1 <= 0:
            raise ValueError("L1 must be greater than zero.")


@dataclass
class Data:
    time: npt.NDArray[np.float64] = field(
        default_factory=lambda: np.array([])
    )
    strain: npt.NDArray[np.float64] = field(
        default_factory=lambda: np.array([])
    )
    stress: npt.NDArray[np.float64] = field(
        default_factory=lambda: np.array([])
    )


class Model:
    params: dict[str, float] = {}
    diagram = ""
    _test: Test
    data: Data

    def __init__(self):
        self.data = Data()

    def __getitem__(self, key: str):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(
                f"The attribute '{key}' does not exist in this model."
            )

    def __str__(self) -> str:
        return self.diagram + "\n" + str(self.params)

    def G(self, t: float) -> float:
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement relaxation modulus."
        )

    def J(self, t: float) -> float:
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement creep compliance."
        )

    def set_time(
        self,
        D1: float | None = None,
        L1: float | None = None,
        D2: float | None = None,
        L2: float | None = None,
        D1_size: int = 21,
        L1_size: int = 21,
        D2_size: int = 21,
        L2_size: int = 21,
    ) -> None:
        D1 = self.test.D1 if D1 is None else D1
        L1 = self.test.L1 if L1 is None else L1
        D2 = self.test.D2 if D2 is None else D2
        L2 = self.test.L2 if L2 is None else L2
        t1 = np.linspace(0, D1, D1_size, endpoint=True)
        t2 = D1 + np.linspace(0, L1, L1_size, endpoint=True)
        if D2 == 0:
            t3 = np.array([])
            t4 = np.array([])
        else:
            t3 = D1 + L1 + np.linspace(0, D2, D2_size, endpoint=True)
            if L2 != 0:
                t4 = D1 + L1 + D2 + np.linspace(0, L2, L2_size, endpoint=True)
        self.data.time = np.unique(
            np.concatenate((t1, t2, t3, t4), dtype=float)
        )

    def set_test(self, test: Test) -> None:
        self.test = test

    @property
    def test(self) -> Test:
        return self._test

    @test.setter
    def test(self, test: Test) -> None:
        if not isinstance(test, Test):
            raise ValueError("test is not a Test object.")

        self._test = test

    def time_between(self, first: float, last: float) -> npt.NDArray[np.float64]:
        time = self.data.time
        return time[(first <= time) & (time <= last)]

    def _input(self) -> npt.NDArray[np.float64]:
        input_D1 = self.time_between(
            0, self.test.D1) * self.test.I / self.test.D1
        input_L1 = self.test.I * \
            np.ones_like(self.time_between(
                self.test.D1, self.test.L1 + self.test.D1))
        if self.test.D2 == 0:
            input_D2 = np.array([])
            input_L2 = np.array([])
        else:
            first = self.test.L1 + self.test.D1
            last = self.test.L1 + self.test.D1 + self.test.D2
            input_D2 = self.test.I * \
                (1 - (self.time_between(first, last) -
                 self.test.L1 - self.test.D1) / self.test.D2)
            if self.test.L2 != 0:
                first = self.test.L1 + self.test.D1 + self.test.D2
                last = self.test.L1 + self.test.D1 + self.test.D2 + self.test.L2
                input_L2 = np.zeros_like(self.time_between(first, last))
        return np.concatenate((input_D1, input_L1[1:], input_D2[1:], input_L2[1:]), dtype=float)

    @lru_cache(maxsize=100)
    def _input_rate(self, t: float) -> float:
        if t < self.test.D1:
            return self.test.I / self.test.D1
        if self.test.L1 + self.test.D1 <= t < self.test.D1 + self.test.L1 + self.test.D2:
            return - self.test.I / self.test.D2
        return 0.0

    def set_input(self) -> None:
        if self.test.method == TestMethod.CREEP:
            self.data.stress = self._input()
        if self.test.method == TestMethod.RELAXATION:
            self.data.strain = self._input()

    @staticmethod
    @lru_cache(maxsize=100)
    def quad(func: Callable, a: float, b: float) -> float:
        return quad(func, a, b, epsrel=1e-3)[0]

    def get_output(self, t: float) -> float:
        if self.test.method == TestMethod.CREEP:
            func = self.J
        if self.test.method == TestMethod.RELAXATION:
            func = self.G

        def _integrand(t_prime: float) -> float:
            return func(t - t_prime) * self._input_rate(t_prime)

        def _boltzmann_integral(t: float) -> float:
            t = round(t, max(0, int(-np.log10(self.test.D1))) + 4)
            if t < 0:
                raise ValueError("Invalid time value.")
            if t == 0:
                return 0.0
            if t < self.test.D1:
                return Model.quad(_integrand, 0, t)
            else:
                p1 = Model.quad(_integrand, 0, self.test.D1)
                if t < self.test.D1 + self.test.L1:
                    return p1
                else:
                    if t < self.test.D1 + self.test.L1 + self.test.D2:
                        return p1 + Model.quad(_integrand, self.test.L1 + self.test.D1, t)
                    else:
                        p2 = Model.quad(
                            _integrand,
                            self.test.L1 + self.test.D1,
                            self.test.D1 + self.test.L1 + self.test.D2,
                        )
                        return p1 + p2
        return _boltzmann_integral(t)

    def run(self) -> None:
        if len(self.data.time) == 0:
            raise ValueError("Set model time before the run.")
        if self.test.method == TestMethod.CREEP:
            self.data.strain = np.vectorize(self.get_output)(self.data.time)
        if self.test.method == TestMethod.RELAXATION:
            self.data.stress = np.vectorize(self.get_output)(self.data.time)

    def get_data(self, first: float, last: float) -> Data:
        idx1 = np.argmin(np.abs(self.data.time - first))
        idx2 = np.argmin(np.abs(self.data.time - last))
        return Data(
            time=self.data.time[idx1:idx2+1],
            strain=self.data.strain[idx1:idx2+1],
            stress=self.data.stress[idx1:idx2+1]
        )

    # def trigger_force(self, trigger: float) -> None:
    #     while True:
    #         self.run()
    #         approach = self.get_approach()
    #         stress = approach["stress"]
    #         max_stress = stress[-1]
    #         if max_stress < 0:
    #             self.stress = -1 * self.time
    #             break
    #         if np.isclose(max_stress, trigger, rtol=1e-3):
    #             break
    #         if max_stress > trigger:
    #             rate = self.I / self.D
    #             trigger_idx = np.argmin(np.abs(stress - trigger)).astype(int)
    #             trigger_idx = int(max(trigger_idx, 1))
    #             self.D = self.time[trigger_idx]
    #             self.I = self.D * rate
    #         elif max_stress < trigger:
    #             scale = trigger / max_stress
    #             self.I *= scale
    #             self.D *= scale



import warnings

import numpy as np

from scipy.special import gamma, gammaln


class MittagLeffler:
    z_thresh: float | None = None
    thresh_eval: float = np.inf

    def __init__(self) -> None:
        pass

    def __call__(self, a: float, b: float, z: float) -> float:
        if self.z_thresh is None:
            eval_current = mittleff(a, b, z)
            if abs(eval_current) <= self.thresh_eval:
                self.thresh_eval = eval_current
            else:
                self.z_thresh = z

        return self.thresh_eval


def mittleff(
    a: float,
    b: float,
    z: float,
    max_terms: int = 1000,
    tol: float = 1e-5
) -> float:
    result = 0.0
    abs_z = abs(z)
    if abs_z < 1.e-15:
        if b > 0:
            return 1.0 / gamma(b)
        else:
            warnings.warn(
                "Invalid parameters: log(0) encountered.", RuntimeWarning
            )
            return np.nan
    sign_z = np.sign(z)
    for n in range(max_terms):
        log_term = n * np.log(abs_z) - gammaln(a * n + b)
        term = np.exp(log_term) * sign_z**n
        result += term
        if abs(term) < tol * abs(result):
            break

    if n == max_terms - 1:
        warnings.warn(
            f"_ml failed to converge with {n} terms. Increase max_terms.", RuntimeWarning
        )

    return result



import unittest

import numpy as np

from pyvisq import Test, TestMethod
from pyvisq.models import Model


class TestModel(Model):
    def G(self, t: float) -> float:
        return 1.0

    def J(self, t: float) -> float:
        return 1.0


I = 1.0
D1 = 0.1
L1 = 0.2
D2 = 0.3
L2 = 0.4
creep = Test(method=TestMethod.CREEP, I=I, D1=D1, L1=L1, D2=D2, L2=L2)
relaxation = Test(method=TestMethod.RELAXATION,
                  I=I, D1=D1, L1=L1, D2=D2, L2=L2)
model = TestModel()


class TestModelMethodsCreep(unittest.TestCase):

    def setUp(self):
        self.model = model
        self.model.set_test(creep)
        self.model.set_time()

    def test_set_time(self):
        self.model.set_time()
        time = self.model.data.time
        self.assertEqual(len(time), 401)
        self.assertAlmostEqual(time[-1], 1.0)

    def test_set_test(self):
        self.model.set_test(creep)
        self.assertEqual(self.model.test.I, 1.0)
        self.assertEqual(self.model.test.D1,  0.1)
        self.assertEqual(self.model.test.L1,  0.2)
        self.assertEqual(self.model.test.D2,  0.3)
        self.assertEqual(self.model.test.L2,  0.4)

    def test_input1(self):
        self.model.set_test(creep)
        self.model.set_time()
        input = self.model._input()
        self.assertEqual(len(input), 401)
        self.assertAlmostEqual(input[-1], 0.0)

    def test_input2(self):
        self.model.set_test(creep)
        self.model.set_time()
        input = self.model._input()
        self.assertEqual(len(input), 401)
        self.assertAlmostEqual(input[101], 1.0)


if __name__ == '__main__':
    unittest.main()



import numpy as np

from functools import lru_cache
from dataclasses import dataclass

from .model import Model
from .maxwell import Maxwell, MaxwellParams
from .kelvinvoigt import KelvinVoigt, KelvinVoigtParams
from .elements import Spring, Dashpot, SpringParams, DashpotParams


@dataclass
class BurgersParams():
    dashpot_a: DashpotParams
    spring_a: SpringParams
    dashpot_b: DashpotParams
    spring_b: SpringParams


class Burgers(Model):
    _params: BurgersParams
    dashpot_a: Dashpot
    spring_a: Spring
    maxwell_branch: Maxwell
    dashpot_b: Dashpot
    spring_b: Spring
    kelvinvoigt_branch: KelvinVoigt
    diagram = """
                                                        ___
                                                _________| |________
                ___                            |        _|_|  cb    |
            _____| |________╱╲  ╱╲  ╱╲  _______|                    |____
                _|_|  ca      ╲╱  ╲╱  ╲╱  ka   |                    |
                                               |____╱╲  ╱╲  ╱╲  ____|
                                                      ╲╱  ╲╱  ╲╱  kb
            """

    def __init__(self, params: BurgersParams) -> None:
        super().__init__()
        self.params = params
        self.maxwell_branch = Maxwell(
            MaxwellParams(params.dashpot_a, params.spring_a)
        )
        self.dashpot_a = self.maxwell_branch.dashpot
        self.spring_a = self.maxwell_branch.spring
        self.kelvinvoigt_branch = KelvinVoigt(
            KelvinVoigtParams(params.dashpot_a, params.spring_a)
        )
        self.dashpot_b = self.kelvinvoigt_branch.dashpot
        self.spring_b = self.kelvinvoigt_branch.spring

    @property
    def params(self) -> BurgersParams:
        return self._params

    @params.setter
    def params(self, params: BurgersParams) -> None:
        if not isinstance(params, BurgersParams):
            raise ValueError("Invalid parameters: Expected PowerlawParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        ca = self.params.dashpot_a.c
        ka = self.params.spring_a.k
        cb = self.params.dashpot_b.c
        kb = self.params.spring_b.k
        p1 = ca/ka + ca/kb + cb/kb
        p2 = ca * cb / (ka * kb)
        q1 = ca
        q2 = ca * cb / kb
        A = np.sqrt(p1**2 - 4*p2)
        r1 = (p1 - A) / (2 * p2)
        r2 = (p1 + A) / (2 * p2)
        return ((q1 - q2 * r1) * np.exp(-r1 * t) - (q1 - q2 * r2) * np.exp(-r2 * t)) / A

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        return self.spring_a.J(t) + self.dashpot_a.J(t) + self.kelvinvoigt_branch.J(t)



import numpy as np

from scipy.special import gamma
from functools import lru_cache
from dataclasses import dataclass

from .model import Model


@dataclass
class SpringParams():
    k: float

    def __post_init__(self):
        if self.k <= 0:
            raise ValueError("k must be greater than zero.")


class Spring(Model):
    _params: SpringParams
    diagram = """
                ____╱╲  ╱╲  ╱╲  ____
                      ╲╱  ╲╱  ╲╱  k
                """

    def __init__(self, params: SpringParams) -> None:
        super().__init__()
        self.params = params

    @property
    def params(self) -> SpringParams:
        return self._params

    @params.setter
    def params(self, params: SpringParams) -> None:
        if not isinstance(params, SpringParams):
            raise ValueError("Invalid parameters: Expected SpringParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        k = self.params.k
        return k

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        k = self.params.k
        return 1 / k


@dataclass
class DashpotParams():
    c: float

    def __post_init__(self):
        if self.c <= 0:
            raise ValueError("c must be greater than zero.")


class Dashpot(Model):
    _params: DashpotParams
    diagram = """
                    ___
                _____| |_____
                    _|_|  c
                """

    def __init__(self, params: DashpotParams) -> None:
        super().__init__()
        self.params = params

    @property
    def params(self) -> DashpotParams:
        return self._params

    @params.setter
    def params(self, params: DashpotParams) -> None:
        if not isinstance(params, DashpotParams):
            raise ValueError("Invalid parameters: Expected DashpotParams.")
        self._params = params

    def G(self, t: float) -> float:
        dirac = np.inf if t == 0 else 0
        return self.params.c * dirac

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        c = self.params.c
        return t / c


@dataclass
class SpringpotParams():
    e: float
    ce: float

    def __post_init__(self):
        if not (0 <= self.e <= 1):
            raise ValueError("b must be between 0 and 1.")
        if self.ce <= 0:
            raise ValueError("cb must be greater than zero.")


class Springpot(Model):
    _params: SpringpotParams
    diagram = """
                ____╱╲____
                    ╲╱  ce, e
                """

    def __init__(self, params: SpringpotParams) -> None:
        super().__init__()
        self.params = params

    @property
    def params(self) -> SpringpotParams:
        return self._params

    @params.setter
    def params(self, params: SpringpotParams) -> None:
        if not isinstance(params, SpringpotParams):
            raise ValueError("Invalid parameters: Expected SpringpotParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        t = max(t, 1e-10)
        e = self.params.e
        ce = self.params.ce
        return ce * t ** (-e) / gamma(1 - e)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        e = self.params.e
        ce = self.params.ce
        return (t ** e) / (ce * gamma(1 + e))



# this is python version of poynting_thomson.jl in the folder "pyvisq/models"
import mpmath

import numpy as np

from functools import lru_cache
from dataclasses import dataclass

from .model import Model
from .elements import Spring, Dashpot, Springpot, SpringParams, DashpotParams, SpringpotParams
from .kelvinvoigt import KelvinVoigt, KelvinVoigtParams, FracDashpotKelvinVoigt, FracDashpotKelvinVoigtParams, FracSpringKelvinVoigt, FracSpringKelvinVoigtParams, FracKelvinVoigt, FracKelvinVoigtParams
from ..utils import MittagLeffler


@dataclass
class SLSPTParams():
    dashpot_a: DashpotParams
    spring_b: SpringParams
    spring_c: SpringParams


class SLSPT(Model):
    _params: SLSPTParams
    dashpot_a: Dashpot
    spring_b: Spring
    kelvingvoigt_branch: KelvinVoigt
    spring_c: Spring
    diagram = """
                         ___
                 _________| |________
                |        _|_|   ca   |
            ____|                    |______╱╲  ╱╲  ╱╲  ____
                |                    |        ╲╱  ╲╱  ╲╱  kc
                |____╱╲  ╱╲  ╱╲  ____|
                       ╲╱  ╲╱  ╲╱  kb
            """

    def __init__(self, params: SLSPTParams) -> None:
        super().__init__()
        self.params = params
        self.kelvingvoigt_branch = KelvinVoigt(
            KelvinVoigtParams(params.dashpot_a, params.spring_b)
        )
        self.dashpot_a = self.kelvingvoigt_branch.dashpot
        self.spring_b = self.kelvingvoigt_branch.spring
        self.spring_c = Spring(params.spring_c)

    @property
    def params(self) -> SLSPTParams:
        return self._params

    @params.setter
    def params(self, params: SLSPTParams) -> None:
        if not isinstance(params, SLSPTParams):
            raise ValueError("Invalid parameters: Expected SLSPTParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        ca = self.params.dashpot_a.c
        kb = self.params.spring_b.k
        kc = self.params.spring_c.k

        Zkb = (kc**2) / (kb + kc)
        Zca = ca * (kc**2) / (kb + kc)**2
        Zkc = (kb * kc) / (kb + kc)
        return Zkc + Zkb * np.exp(-t * Zkb / Zca)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        return self.kelvingvoigt_branch.J(t) + self.spring_c.J(t)


@dataclass
class JeffreysPTParams():
    dashpot_a: DashpotParams
    spring_b: SpringParams
    dashpot_c: DashpotParams


class JeffreysPT(Model):
    _params: JeffreysPTParams
    dashpot_a: Dashpot
    spring_b: Spring
    kelvingvoigt_branch: KelvinVoigt
    dashpot_c: Dashpot
    diagram = """
                         ___
                 _________| |________
                |        _|_|  ca    |      __
            ____|                    |_______| |____
                |                    |      _|_|  cc
                |____╱╲  ╱╲  ╱╲  ____|
                       ╲╱  ╲╱  ╲╱  kb
            """

    def __init__(self, params: JeffreysPTParams) -> None:
        super().__init__()
        self.params = params
        self.kelvingvoigt_branch = KelvinVoigt(
            KelvinVoigtParams(params.dashpot_a, params.spring_b)
        )
        self.dashpot_a = self.kelvingvoigt_branch.dashpot
        self.spring_b = self.kelvingvoigt_branch.spring
        self.dashpot_c = Dashpot(params.dashpot_c)

    @property
    def params(self) -> JeffreysPTParams:
        return self._params

    @params.setter
    def params(self, params: JeffreysPTParams) -> None:
        if not isinstance(params, JeffreysPTParams):
            raise ValueError("Invalid parameters: Expected JeffreysPTParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        ca = self.params.dashpot_a.c
        kb = self.params.spring_b.k
        cc = self.params.dashpot_c.c

        Zca = (cc**2) / (ca + cc)
        Zkb = kb * (cc**2) / (ca + cc)**2
        Zcc = (ca * cc) / (ca + cc)
        diracterm = 0.0 if t != 0.0 else np.inf
        return Zkb * np.exp(-Zkb * t / Zca) + Zcc * diracterm

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        return self.kelvingvoigt_branch.J(t) + self.dashpot_c.J(t)


@dataclass
class FracSLSPTParams():
    springpot_a: SpringpotParams
    spring_b: SpringParams
    spring_c: SpringParams


class FracSLSPT(Model):
    _params: FracSLSPTParams
    springpot_a: Springpot
    spring_b: Spring
    kelvingvoigt_branch: FracSpringKelvinVoigt
    spring_c: Spring
    ml: MittagLeffler
    diagram = """
                 _________╱╲_________
                |         ╲╱  ca, a  |
            ____|                    |______╱╲  ╱╲  ╱╲  ____
                |                    |        ╲╱  ╲╱  ╲╱  kc
                |____╱╲  ╱╲  ╱╲  ____|
                       ╲╱  ╲╱  ╲╱  kb
            """

    def __init__(self, params: FracSLSPTParams) -> None:
        super().__init__()
        self.params = params
        self.kelvingvoigt_branch = FracSpringKelvinVoigt(
            FracSpringKelvinVoigtParams(params.springpot_a, params.spring_b)
        )
        self.springpot_a = self.kelvingvoigt_branch.springpot
        self.spring_b = self.kelvingvoigt_branch.spring
        self.spring_c = Spring(params.spring_c)
        self.ml = MittagLeffler()

    @property
    def params(self) -> FracSLSPTParams:
        return self._params

    @params.setter
    def params(self, params: FracSLSPTParams) -> None:
        if not isinstance(params, FracSLSPTParams):
            raise ValueError("Invalid parameters: Expected FracSLSPTParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        a = self.params.springpot_a.e
        ca = self.params.springpot_a.ce
        kb = self.params.spring_b.k
        kc = self.params.spring_c.k

        Zkb = (kc**2) / (kb + kc)
        Zca = ca * (kc**2) / (kb + kc)**2
        Zkc = (kb * kc) / (kb + kc)
        return Zkb * self.ml(a, a, - (Zkb / Zca) * (t**a)) + Zkc

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        return self.kelvingvoigt_branch.J(t) + self.spring_c.J(t)


@dataclass
class FracJeffreysPTParams():
    dashpot_a: DashpotParams
    springpot_b: SpringpotParams
    dashpot_c: DashpotParams


class FracJeffreysPT(Model):
    _params: FracJeffreysPTParams
    dashpot_a: Dashpot
    springpot_a: Springpot
    kelvingvoigt_branch: FracDashpotKelvinVoigt
    dashpot_c: Dashpot
    ml: MittagLeffler
    diagram = """
                         ___
                 _________| |________
                |        _|_|  ca    |        ___
            ____|                    |_________| |_____
                |                    |        _|_|  cc
                |_________╱╲_________|
                          ╲╱  cb, b
            """

    def __init__(self, params: FracJeffreysPTParams) -> None:
        super().__init__()
        self.params = params
        self.kelvingvoigt_branch = FracDashpotKelvinVoigt(
            FracDashpotKelvinVoigtParams(params.dashpot_a, params.springpot_b)
        )
        self.dashpot_a = self.kelvingvoigt_branch.dashpot
        self.springpot_a = self.kelvingvoigt_branch.springpot
        self.dashpot_c = Dashpot(params.dashpot_c)
        self.ml = MittagLeffler()

    @property
    def params(self) -> FracJeffreysPTParams:
        return self._params

    @params.setter
    def params(self, params: FracJeffreysPTParams) -> None:
        if not isinstance(params, FracJeffreysPTParams):
            raise ValueError(
                "Invalid parameters: Expected FracJeffreysPTParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        t = max(t, 1e-10)
        ca = self.params.dashpot_a.c
        b = self.params.springpot_b.e
        cb = self.params.springpot_b.ce
        cc = self.params.dashpot_c.c

        Zca = (cc**2) / (ca + cc)
        Zcb = cb * (cc**2) / (ca + cc)**2
        Zcc = (ca * cc) / (ca + cc)
        diracterm = 0.0 if t != 0.0 else np.inf
        return Zcb * (t**(-b)) * self.ml(1 - b, 1 - b, -Zcb * t**(1 - b) / Zca) + Zcc * diracterm

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        return self.kelvingvoigt_branch.J(t) + self.dashpot_c.J(t)


@dataclass
class FracPTParams():
    springpot_a: SpringpotParams
    springpot_b: SpringpotParams
    springpot_c: SpringpotParams


class FracPT(Model):
    _params: FracPTParams
    springpot_a: Springpot
    springpot_b: Springpot
    kelvingvoigt_branch: FracKelvinVoigt
    springpot_c: Springpot
    diagram = """
                 _________╱╲_________
                |         ╲╱  ca, a  |
            ____|                    |______╱╲______
                |                    |      ╲╱ cc, c
                |_________╱╲_________|
                          ╲╱  cb, b
            """

    def __init__(self, params: FracPTParams) -> None:
        super().__init__()
        self.params = params
        self.kelvingvoigt_branch = FracKelvinVoigt(
            FracKelvinVoigtParams(params.springpot_a, params.springpot_b)
        )
        self.springpot_a = self.kelvingvoigt_branch.springpot_a
        self.springpot_b = self.kelvingvoigt_branch.springpot_b
        self.springpot_c = Springpot(params.springpot_c)

    @property
    def params(self) -> FracPTParams:
        return self._params

    @params.setter
    def params(self, params: FracPTParams) -> None:
        if not isinstance(params, FracPTParams):
            raise ValueError("Invalid parameters: Expected FracPTParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        if t == 0:
            return 0

        a = self.params.springpot_a.e
        ca = self.params.springpot_a.ce
        b = self.params.springpot_b.e
        cb = self.params.springpot_b.ce
        c = self.params.springpot_c.e
        cc = self.params.springpot_c.ce

        def G_hat(s):
            numerator = cc * s**c * (ca * s**a + cb * s**b)
            denominator = cc * s**c + ca * s**a + cb * s**b
            return (1 / s) * numerator / denominator

        t_mp = mpmath.mpf(t)
        G_t = mpmath.invertlaplace(G_hat, t_mp, method='talbot')
        return float(G_t)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        return self.kelvingvoigt_branch.J(t) + self.springpot_c.J(t)



import numpy as np

from functools import lru_cache
from dataclasses import dataclass

from .model import Model
from .elements import Spring, Dashpot, Springpot, SpringParams, DashpotParams, SpringpotParams
from ..utils import MittagLeffler


@dataclass
class KelvinVoigtParams():
    dashpot: DashpotParams
    spring: SpringParams


class KelvinVoigt(Model):
    _params: KelvinVoigtParams
    dashpot: Dashpot
    spring: Spring
    diagram = """
                        ___
                 ________| |_________
                |       _|_|  c      |
            ____|                    |____
                |                    |
                |____╱╲  ╱╲  ╱╲  ____|
                       ╲╱  ╲╱  ╲╱  k
            """

    def __init__(self, params: KelvinVoigtParams) -> None:
        super().__init__()
        self.params = params
        self.spring = Spring(params.spring)
        self.dashpot = Dashpot(params.dashpot)

    @property
    def params(self) -> KelvinVoigtParams:
        return self._params

    @params.setter
    def params(self, params: KelvinVoigtParams) -> None:
        if not isinstance(params, KelvinVoigtParams):
            raise ValueError("Invalid parameters: Expected KelvinVoigtParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.spring.G(t)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        k = self.params.spring.k
        c = self.params.dashpot.c
        tau = c / k
        return (1 - np.exp(-t / tau)) / k


@dataclass
class FracDashpotKelvinVoigtParams():
    dashpot: DashpotParams
    springpot: SpringpotParams


class FracDashpotKelvinVoigt(Model):
    _params: FracDashpotKelvinVoigtParams
    dashpot: Dashpot
    springpot: Springpot
    ml: MittagLeffler
    diagram = """
                        ___
                 ________| |_________
                |       _|_|  c      |
            ____|                    |____
                |                    |
                |_________╱╲_________|
                          ╲╱  cb, b
            """

    def __init__(self, params: FracDashpotKelvinVoigtParams) -> None:
        super().__init__()
        self.params = params
        self.dashpot = Dashpot(params.dashpot)
        self.springpot = Springpot(params.springpot)
        self.ml = MittagLeffler()

    @property
    def params(self) -> FracDashpotKelvinVoigtParams:
        return self._params

    @params.setter
    def params(self, params: FracDashpotKelvinVoigtParams) -> None:
        if not isinstance(params, FracDashpotKelvinVoigtParams):
            raise ValueError(
                "Invalid parameters: Expected FracDashpotKelvinVoigtParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.springpot.G(t)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        c = self.params.dashpot.c
        b = self.params.springpot.e
        cb = self.params.springpot.ce
        return (t / c) * self.ml(1 - b, 1 + 1, -cb * t ** (1 - b) / c)


@dataclass
class FracSpringKelvinVoigtParams():
    springpot: SpringpotParams
    spring: SpringParams


class FracSpringKelvinVoigt(Model):
    _params: FracSpringKelvinVoigtParams
    springpot: Springpot
    spring: Spring
    ml: MittagLeffler
    diagram = """
                 _________╱╲_________
                |         ╲╱  ca, a  |
            ____|                    |____
                |                    |
                |_____╱╲  ╱╲  ╱╲  ___|
                        ╲╱  ╲╱  ╲╱  k
            """

    def __init__(self, params: FracSpringKelvinVoigtParams) -> None:
        super().__init__()
        self.params = params
        self.spring = Spring(params.spring)
        self.springpot = Springpot(params.springpot)
        self.ml = MittagLeffler()

    @property
    def params(self) -> FracSpringKelvinVoigtParams:
        return self._params

    @params.setter
    def params(self, params: FracSpringKelvinVoigtParams) -> None:
        if not isinstance(params, FracSpringKelvinVoigtParams):
            raise ValueError(
                "Invalid parameters: Expected FracSpringKelvinVoigtParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.spring.G(t) + self.springpot.G(t)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        k = self.params.spring.k
        a = self.params.springpot.e
        ca = self.params.springpot.ce
        return (t ** a / ca) * self.ml(a, 1 + a, -k * t ** a / ca)


@dataclass
class FracKelvinVoigtParams():
    springpot_a: SpringpotParams
    springpot_b: SpringpotParams


class FracKelvinVoigt(Model):
    _params: FracKelvinVoigtParams
    springpot_a: Springpot
    springpot_b: Springpot
    ml: MittagLeffler
    diagram = """
                 _________╱╲_________
                |         ╲╱  ca, a  |
            ____|                    |____
                |                    |
                |_________╱╲_________|
                          ╲╱  cb, b
            """

    def __init__(self, params: FracKelvinVoigtParams) -> None:
        super().__init__()
        self.params = params
        self.springpot_a = Springpot(params.springpot_a)
        self.springpot_b = Springpot(params.springpot_b)
        self.ml = MittagLeffler()

    @property
    def params(self) -> FracKelvinVoigtParams:
        return self._params

    @params.setter
    def params(self, params: FracKelvinVoigtParams) -> None:
        if not isinstance(params, FracKelvinVoigtParams):
            raise ValueError(
                "Invalid parameters: Expected FracKelvinVoigtParams.")
        self._params = params

    @lru_cache(maxsize=100)
    def G(self, t: float) -> float:
        return self.springpot_a.G(t) + self.springpot_b.G(t)

    @lru_cache(maxsize=100)
    def J(self, t: float) -> float:
        a = self.params.springpot_a.e
        ca = self.params.springpot_a.ce
        b = self.params.springpot_b.e
        cb = self.params.springpot_b.ce
        return (t ** a / ca) * self.ml(a - b, 1 + a, -cb * t ** (a - b) / ca)
